/**
 * Copyright Sterfive 2020-2022
 */

import { Guid, UInt32 } from "node-opcua-basic-types";
import { NodeId } from "node-opcua-nodeid";
import { IBasicSession } from "node-opcua-pseudo-session";
import { StatusCode } from "node-opcua-status-code";
import { ConfigurationVersionDataType, DataSetMetaDataType, PublishedVariableDataType, SimpleAttributeOperand, ContentFilter, PublishedVariableDataTypeOptions } from "node-opcua-types";
import { UANodeProxy } from "./ua_node_proxy";
import { DataSetWriter } from "./dataset_writer";
declare const publishedDataSetNodeIds: Record<string, string>;
export declare abstract class PublishedDataSet<T extends typeof publishedDataSetNodeIds> extends UANodeProxy<T> {
    configurationVersion?: ConfigurationVersionDataType | undefined;
    addResults?: StatusCode[] | undefined;
    constructor(session: IBasicSession, nodeId: NodeId, configurationVersion?: ConfigurationVersionDataType | undefined, addResults?: StatusCode[] | undefined);
    getConfigurationVersion(): Promise<ConfigurationVersionDataType>;
    getDataSetMetaData(): Promise<DataSetMetaDataType>;
    getDataSetClassId(): Promise<Guid>;
    getDataSetWriterCollection(): Promise<DataSetWriter[]>;
}
declare const publishedDataItemsNodeIds: Record<string, string>;
export declare class PublishedDataItems extends PublishedDataSet<typeof publishedDataItemsNodeIds> {
    protected _getPaths(): Record<string, string>;
    getPublishedData(): Promise<PublishedVariableDataType[]>;
    /**
     *
     * https://reference.opcfoundation.org/v104/Core/docs/Part14/9.1.4/#9.1.4.2.1
     *
     *
     * [out]	ConfigurationVersionDataType		NewConfigurationVersion
     * [out]  StatusCode[]
     *
     *
     * @param {ConfigurationVersion} configurationVersion  Configuration version of the DataSet.
     * The configuration version and the indices passed in through VariablesToRemove must match the entire current configuration
     * version of the Object when the Method call is processed. If it does not match, the result BadInvalidState shall be returned
     *
     * @param {String[]} fieldNameAliases
     * The names assigned to the selected Variables for the fields in the DataSetMetaData and in the DataSetMessages
     * for tagged message encoding. The size and the order of the array shall match the VariablesToAdd.
     * The string shall be used to set the name field in the FieldMetaData that is part of the DataSetMetaData.
     *
     * @param {boolean[]} promotedFields
     * The flags indicating if the corresponding field is promoted to the DataSetMessage header. The size and the
     * order of the array shall match the VariablesToAdd. The flag is used to set the PromotedField flag in the fieldFlags
     *  parameter in the FieldMetaData.
     *
     * @param {PublishedVariableDataType[]} variablesToAdd
     * Array of Variables to add to PublishedData and the related configuration settings. Successfully added variables are appended
     * to the end of the list of published variables configured in the PublishedData Property. Failed variables are not added to
     * the list.
     * The parameters builtInType, dataType, valueRank and arrayDimensions of the FieldMetaData are filled from corresponding Variable Attributes
     */
    addVariables({ configurationVersion, fieldNameAliases, promotedFields, variablesToAdd, }: {
        configurationVersion: ConfigurationVersionDataType;
        fieldNameAliases: string[];
        promotedFields: boolean[];
        variablesToAdd: PublishedVariableDataTypeOptions[];
    }): Promise<{
        newConfiguration: ConfigurationVersionDataType;
        statusCodes: StatusCode[];
    }>;
    /**
     * This Method is used to remove Variables from the PublishedData list.
     * It contains the list of published Variables of a PublishedDataItemsType Object.
     *
     * A caller shall read the current Values of PublishedData and ConfigurationVersion prior
     * to calling this Method, to ensure the use of the correct index of the Variables that are being removed.
     *
     * @param configurationVersion: Configuration version of the DataSet.
     * The configuration version and the indices passed in through
     * VariablesToRemove must match the entire current configuration version of the Object when the Method
     * call is processed. If it does not match, the result Bad_InvalidState shall be returned
     *
     * @param VariablesToRemove:
     * Array of indices of Variables to remove from the list of Variables configured in PublishedData of
     *   the PublishedDataItemsType. This matches the list of fields configured in the DataSetMetaData
     * of the PublishedDataSetType.
     */
    removeVariables({ configurationVersion, variablesToRemove, }: {
        configurationVersion: ConfigurationVersionDataType;
        variablesToRemove: UInt32[];
    }): Promise<{
        newConfiguration: ConfigurationVersionDataType;
        removeResults: StatusCode[];
    }>;
}
declare const PublishedEventsNodeIds: Record<string, string>;
export declare class PublishedEvents extends PublishedDataSet<typeof PublishedEventsNodeIds> {
    protected _getPaths(): Record<string, string>;
    getEventNotifier(): Promise<NodeId>;
    getSelectedFields(): Promise<SimpleAttributeOperand[]>;
    getFilter(): Promise<ContentFilter>;
    modifyFieldSelection(): Promise<void>;
}
export {};
